#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <stdio.h>
#include <tchar.h>
#include <cfgmgr32.h>
#include <newdev.h>
#include "resource.h"
#include "msg.h"

typedef int (*CallbackFunc)(HDEVINFO Devs,PSP_DEVINFO_DATA DevInfo,DWORD Index,LPVOID Context);
int EnumerateDevices(DWORD Flags,LPTSTR strClass,LPTSTR strDevice,CallbackFunc Callback,LPVOID Context);
struct IdEntry {
	LPCTSTR String;     // string looking for
	LPCTSTR Wild;       // first wild character if any
	BOOL    InstanceId;
};
#define INSTANCEID_PREFIX_CHAR TEXT('@') // character used to prefix instance ID's
#define WILD_CHAR              TEXT('*') // wild character
#define QUOTE_PREFIX_CHAR      TEXT('\'') // prefix character to ignore wild characters

//
// exit codes
//
#define EXIT_OK      (0)
#define EXIT_REBOOT  (1)
#define EXIT_FAIL    (2)
#define EXIT_USAGE   (3)

typedef struct {
	DWORD count;
	DWORD control;
	BOOL  reboot;
	LPCTSTR strSuccess;
	LPCTSTR strReboot;
	LPCTSTR strFail;
}
GenericContext;




void FormatToStream(FILE * stream,DWORD fmt,...)
/*++

Routine Description:

Format text to stream using a particular msg-id fmt
Used for displaying localizable messages

Arguments:

stream              - file stream to output to, stdout or stderr
fmt                 - message id
...                 - parameters %1...

Return Value:

none

--*/
{
	va_list arglist;
	LPTSTR locbuffer = NULL;
	DWORD count;

	va_start(arglist, fmt);
	count = FormatMessage(FORMAT_MESSAGE_FROM_HMODULE|FORMAT_MESSAGE_ALLOCATE_BUFFER,
		NULL,
		fmt,
		0,              // LANGID
		(LPTSTR) &locbuffer,
		0,              // minimum size of buffer
		&arglist);

	if(locbuffer) {
		if(count) {
			_fputts(locbuffer,stream);
		}
		LocalFree(locbuffer);
	}
}

void Padding(int pad)
/*++

Routine Description:

Insert padding into line before text

Arguments:

pad - number of padding tabs to insert

Return Value:

none

--*/
{
	int c;

	for(c=0;c<pad;c++) {
		fputs("    ",stdout);
	}
}


void Usage(LPCTSTR BaseName)
/*++

Routine Description:

Display simple usage text

Arguments:

BaseName            - name of executable

Return Value:

none

--*/
{
	FormatToStream(stderr,MSG_USAGE,BaseName);
}

void CommandUsage(LPCTSTR BaseName,LPCTSTR Cmd)
/*++

Routine Description:

Invalid command usage
Display how to get help on command

Arguments:

BaseName            - name of executable

Return Value:

none

--*/
{
	FormatToStream(stderr,MSG_COMMAND_USAGE,BaseName,Cmd);
}

void Failure(LPCTSTR BaseName,LPCTSTR Cmd)
/*++

Routine Description:

Display simple error text for general failure

Arguments:

BaseName            - name of executable

Return Value:

none

--*/
{
	FormatToStream(stderr,MSG_FAILURE,BaseName,Cmd);
}

BOOL Reboot()
/*++

Routine Description:

Attempt to reboot computer

Arguments:

none

Return Value:

TRUE if API suceeded

--*/
{
	HANDLE Token;
	TOKEN_PRIVILEGES NewPrivileges;
	LUID Luid;

	//
	// we need to "turn on" reboot privilege
	// if any of this fails, try reboot anyway
	//
	if(!OpenProcessToken(GetCurrentProcess(),TOKEN_ADJUST_PRIVILEGES,&Token)) {
		goto final;
	}

	if(!LookupPrivilegeValue(NULL,SE_SHUTDOWN_NAME,&Luid)) {
		CloseHandle(Token);
		goto final;
	}

	NewPrivileges.PrivilegeCount = 1;
	NewPrivileges.Privileges[0].Luid = Luid;
	NewPrivileges.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED;

	AdjustTokenPrivileges(
		Token,
		FALSE,
		&NewPrivileges,
		0,
		NULL,
		NULL
		);

	CloseHandle(Token);

final:

	//
	// attempt reboot - inform system that this is planned hardware install
	//
	return ExitWindowsEx(EWX_REBOOT, REASON_PLANNED_FLAG|REASON_HWINSTALL);
}

LPTSTR GetDeviceStringProperty(HDEVINFO Devs,PSP_DEVINFO_DATA DevInfo,DWORD Prop)
/*++

Routine Description:

Return a string property for a device, otherwise NULL

Arguments:

Devs    )_ uniquely identify device
DevInfo )
Prop     - string property to obtain

Return Value:

string containing description

--*/
{
	LPTSTR buffer;
	DWORD size;
	DWORD reqSize;
	DWORD dataType;
	DWORD szChars;

	size = 1024; // initial guess
	buffer = new TCHAR[(size/sizeof(TCHAR))+1];
	if(!buffer) {
		return NULL;
	}
	while(!SetupDiGetDeviceRegistryProperty(Devs,DevInfo,Prop,&dataType,(LPBYTE)buffer,size,&reqSize)) {
		if(GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
			goto failed;
		}
		if(dataType != REG_SZ) {
			goto failed;
		}
		size = reqSize;
		delete [] buffer;
		buffer = new TCHAR[(size/sizeof(TCHAR))+1];
		if(!buffer) {
			goto failed;
		}
	}
	szChars = reqSize/sizeof(TCHAR);
	buffer[szChars] = TEXT('\0');
	return buffer;

failed:
	if(buffer) {
		delete [] buffer;
	}
	return NULL;
}

LPTSTR GetDeviceDescription(HDEVINFO Devs,PSP_DEVINFO_DATA DevInfo)
/*++

Routine Description:

Return a string containing a description of the device, otherwise NULL
Always try friendly name first

Arguments:

Devs    )_ uniquely identify device
DevInfo )

Return Value:

string containing description

--*/
{
	LPTSTR desc;
	desc = GetDeviceStringProperty(Devs,DevInfo,SPDRP_FRIENDLYNAME);
	if(!desc) {
		desc = GetDeviceStringProperty(Devs,DevInfo,SPDRP_DEVICEDESC);
	}
	return desc;
}

IdEntry GetIdType(LPCTSTR Id)
/*++

Routine Description:

Determine if this is instance id or hardware id and if there's any wildcards
instance ID is prefixed by '@'
wildcards are '*'


Arguments:

Id - ptr to string to check

Return Value:

IdEntry

--*/
{
	IdEntry Entry;

	Entry.InstanceId = FALSE;
	Entry.Wild = NULL;
	Entry.String = Id;

	if(Entry.String[0] == INSTANCEID_PREFIX_CHAR) {
		Entry.InstanceId = TRUE;
		Entry.String = CharNext(Entry.String);
	}
	if(Entry.String[0] == QUOTE_PREFIX_CHAR) {
		//
		// prefix to treat rest of string literally
		//
		Entry.String = CharNext(Entry.String);
	} else {
		//
		// see if any wild characters exist
		//
		Entry.Wild = _tcschr(Entry.String,WILD_CHAR);
	}
	return Entry;
}

LPTSTR * GetMultiSzIndexArray(LPTSTR MultiSz)
/*++

Routine Description:

Get an index array pointing to the MultiSz passed in

Arguments:

MultiSz - well formed multi-sz string

Return Value:

array of strings. last entry+1 of array contains NULL
returns NULL on failure

--*/
{
	LPTSTR scan;
	LPTSTR * array;
	int elements;

	for(scan = MultiSz, elements = 0; scan[0] ;elements++) {
		scan += lstrlen(scan)+1;
	}
	array = new LPTSTR[elements+2];
	if(!array) {
		return NULL;
	}
	array[0] = MultiSz;
	array++;
	if(elements) {
		for(scan = MultiSz, elements = 0; scan[0]; elements++) {
			array[elements] = scan;
			scan += lstrlen(scan)+1;
		}
	}
	array[elements] = NULL;
	return array;
}

void DelMultiSz(LPTSTR * Array)
/*++

Routine Description:

Deletes the string array allocated by GetDevMultiSz/GetRegMultiSz/GetMultiSzIndexArray

Arguments:

Array - pointer returned by GetMultiSzIndexArray

Return Value:

None

--*/
{
	if(Array) {
		Array--;
		if(Array[0]) {
			delete [] Array[0];
		}
		delete [] Array;
	}
}

LPTSTR * GetDevMultiSz(HDEVINFO Devs,PSP_DEVINFO_DATA DevInfo,DWORD Prop)
/*++

Routine Description:

Get a multi-sz device property
and return as an array of strings

Arguments:

Devs    - HDEVINFO containing DevInfo
DevInfo - Specific device
Prop    - SPDRP_HARDWAREID or SPDRP_COMPATIBLEIDS

Return Value:

array of strings. last entry+1 of array contains NULL
returns NULL on failure

--*/
{
	LPTSTR buffer;
	DWORD size;
	DWORD reqSize;
	DWORD dataType;
	LPTSTR * array;
	DWORD szChars;

	size = 8192; // initial guess, nothing magic about this
	buffer = new TCHAR[(size/sizeof(TCHAR))+2];
	if(!buffer) {
		return NULL;
	}
	while(!SetupDiGetDeviceRegistryProperty(Devs,DevInfo,Prop,&dataType,(LPBYTE)buffer,size,&reqSize)) {
		if(GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
			goto failed;
		}
		if(dataType != REG_MULTI_SZ) {
			goto failed;
		}
		size = reqSize;
		delete [] buffer;
		buffer = new TCHAR[(size/sizeof(TCHAR))+2];
		if(!buffer) {
			goto failed;
		}
	}
	szChars = reqSize/sizeof(TCHAR);
	buffer[szChars] = TEXT('\0');
	buffer[szChars+1] = TEXT('\0');
	array = GetMultiSzIndexArray(buffer);
	if(array) {
		return array;
	}

failed:
	if(buffer) {
		delete [] buffer;
	}
	return NULL;
}

LPTSTR * GetRegMultiSz(HKEY hKey,LPCTSTR Val)
/*++

Routine Description:

Get a multi-sz from registry
and return as an array of strings

Arguments:

hKey    - Registry Key
Val     - Value to query

Return Value:

array of strings. last entry+1 of array contains NULL
returns NULL on failure

--*/
{
	LPTSTR buffer;
	DWORD size;
	DWORD reqSize;
	DWORD dataType;
	LPTSTR * array;
	DWORD szChars;
	LONG regErr;

	size = 8192; // initial guess, nothing magic about this
	buffer = new TCHAR[(size/sizeof(TCHAR))+2];
	if(!buffer) {
		return NULL;
	}
	reqSize = size;
	while((regErr = RegQueryValueEx(hKey,Val,NULL,&dataType,(PBYTE)buffer,&reqSize) != NO_ERROR)) {
		if(GetLastError() != ERROR_MORE_DATA) {
			goto failed;
		}
		if(dataType != REG_MULTI_SZ) {
			goto failed;
		}
		size = reqSize;
		delete [] buffer;
		buffer = new TCHAR[(size/sizeof(TCHAR))+2];
		if(!buffer) {
			goto failed;
		}
	}
	szChars = reqSize/sizeof(TCHAR);
	buffer[szChars] = TEXT('\0');
	buffer[szChars+1] = TEXT('\0');

	array = GetMultiSzIndexArray(buffer);
	if(array) {
		return array;
	}

failed:
	if(buffer) {
		delete [] buffer;
	}
	return NULL;
}

BOOL WildCardMatch(LPCTSTR Item,const IdEntry & MatchEntry)
/*++

Routine Description:

Compare a single item against wildcard
I'm sure there's better ways of implementing this
Other than a command-line management tools
it's a bad idea to use wildcards as it implies
assumptions about the hardware/instance ID
eg, it might be tempting to enumerate root\* to
find all root devices, however there is a CfgMgr
API to query status and determine if a device is
root enumerated, which doesn't rely on implementation
details.

Arguments:

Item - item to find match for eg a\abcd\c
MatchEntry - eg *\*bc*\*

Return Value:

TRUE if any match, otherwise FALSE

--*/
{
	LPCTSTR scanItem;
	LPCTSTR wildMark;
	LPCTSTR nextWild;
	size_t matchlen;

	//
	// before attempting anything else
	// try and compare everything up to first wild
	//
	if(!MatchEntry.Wild) {
		return _tcsicmp(Item,MatchEntry.String) ? FALSE : TRUE;
	}
	if(_tcsnicmp(Item,MatchEntry.String,MatchEntry.Wild-MatchEntry.String) != 0) {
		return FALSE;
	}
	wildMark = MatchEntry.Wild;
	scanItem = Item + (MatchEntry.Wild-MatchEntry.String);

	for(;wildMark[0];) {
		//
		// if we get here, we're either at or past a wildcard
		//
		if(wildMark[0] == WILD_CHAR) {
			//
			// so skip wild chars
			//
			wildMark = CharNext(wildMark);
			continue;
		}
		//
		// find next wild-card
		//
		nextWild = _tcschr(wildMark,WILD_CHAR);
		if(nextWild) {
			//
			// substring
			//
			matchlen = nextWild-wildMark;
		} else {
			//
			// last portion of match
			//
			size_t scanlen = lstrlen(scanItem);
			matchlen = lstrlen(wildMark);
			if(scanlen < matchlen) {
				return FALSE;
			}
			return _tcsicmp(scanItem+scanlen-matchlen,wildMark) ? FALSE : TRUE;
		}
		if(_istalpha(wildMark[0])) {
			//
			// scan for either lower or uppercase version of first character
			//
			TCHAR u = _totupper(wildMark[0]);
			TCHAR l = _totlower(wildMark[0]);
			while(scanItem[0] && scanItem[0]!=u && scanItem[0]!=l) {
				scanItem = CharNext(scanItem);
			}
			if(!scanItem[0]) {
				//
				// ran out of string
				//
				return FALSE;
			}
		} else {
			//
			// scan for first character (no case)
			//
			scanItem = _tcschr(scanItem,wildMark[0]);
			if(!scanItem) {
				//
				// ran out of string
				//
				return FALSE;
			}
		}
		//
		// try and match the sub-string at wildMark against scanItem
		//
		if(_tcsnicmp(scanItem,wildMark,matchlen)!=0) {
			//
			// nope, try again
			//
			scanItem = CharNext(scanItem);
			continue;
		}
		//
		// substring matched
		//
		scanItem += matchlen;
		wildMark += matchlen;
	}
	return (wildMark[0] ? FALSE : TRUE);
}

BOOL WildCompareHwIds(LPTSTR * Array,const IdEntry & MatchEntry)
/*++

Routine Description:

Compares all strings in Array against Id
Use WildCardMatch to do real compare

Arguments:

Array - pointer returned by GetDevMultiSz
MatchEntry - string to compare against

Return Value:

TRUE if any match, otherwise FALSE

--*/
{
	if(Array) {
		while(Array[0]) {
			if(WildCardMatch(Array[0],MatchEntry)) {
				return TRUE;
			}
			Array++;
		}
	}
	return FALSE;
}

int EnumerateDevices(DWORD Flags,LPTSTR strClass,LPTSTR strDevice,CallbackFunc Callback,LPVOID Context)
/*++

Routine Description:

Generic enumerator for devices that will be passed the following arguments:
<class> <id>
where <id> can either be @instance-id, or hardware-id and may contain wildcards
<class> is a class name

Arguments:

Flags     - extra enumeration flags (eg DIGCF_PRESENT)
strClass  - class name
strDevice - wildcard string of device id
Callback  - function to call for each hit
Context   - data to pass function for each hit

Return Value:

EXIT_xxxx

--*/
{
	HDEVINFO devs = INVALID_HANDLE_VALUE;
	int failcode = EXIT_FAIL;
	int retcode;
	DWORD devIndex;
	SP_DEVINFO_DATA devInfo;
	SP_DEVINFO_LIST_DETAIL_DATA devInfoListDetail;
	BOOL doSearch = FALSE;
	BOOL match = FALSE;
	GUID cls;
	DWORD numClass = 0;

	//
	// determine the instance id
	//
	IdEntry idDevice = GetIdType(strDevice);
	if(idDevice.Wild || !idDevice.InstanceId) {
		//
		// anything other than simple InstanceId's require a search
		//
		doSearch = TRUE;
	}

	//
	// determine the class
	//
	if(!SetupDiClassGuidsFromNameEx(strClass,&cls,1,&numClass,NULL,NULL) &&
		GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
			goto final;
	}
	if(!numClass) {
		failcode = EXIT_OK;
		goto final;
	}

	if(doSearch) {
		//
		// add all id's to list
		// if there's a class, filter on specified class
		//
		devs = SetupDiGetClassDevsEx(numClass ? &cls : NULL,
			NULL,
			NULL,
			(numClass ? 0 : DIGCF_ALLCLASSES) | Flags,
			NULL,
			NULL,
			NULL);

	} else {
		//
		// blank list, we'll add instance id's by hand
		//
		devs = SetupDiCreateDeviceInfoListEx(numClass ? &cls : NULL,
			NULL,
			NULL,
			NULL);
	}
	if(devs == INVALID_HANDLE_VALUE) {
		goto final;
	}
	if(idDevice.InstanceId) {
		SetupDiOpenDeviceInfo(devs,idDevice.String,NULL,0,NULL);
	}

	devInfoListDetail.cbSize = sizeof(devInfoListDetail);
	if(!SetupDiGetDeviceInfoListDetail(devs,&devInfoListDetail)) {
		goto final;
	}

	devInfo.cbSize = sizeof(devInfo);
	for(devIndex=0;SetupDiEnumDeviceInfo(devs,devIndex,&devInfo);devIndex++) {

		if(doSearch) {
			/*while(!match)*/ {
				TCHAR devID[MAX_DEVICE_ID_LEN];
				LPTSTR *hwIds = NULL;
				LPTSTR *compatIds = NULL;
				//
				// determine instance ID
				//
				if(CM_Get_Device_ID_Ex(devInfo.DevInst,devID,MAX_DEVICE_ID_LEN,0,devInfoListDetail.RemoteMachineHandle)!=CR_SUCCESS) {
					devID[0] = TEXT('\0');
				}

				if(idDevice.InstanceId) {
					//
					// match on the instance ID
					//
					if(WildCardMatch(devID,idDevice)) {
						match = TRUE;
					}
				} else {
					//
					// determine hardware ID's
					// and search for matches
					//
					hwIds = GetDevMultiSz(devs,&devInfo,SPDRP_HARDWAREID);
					compatIds = GetDevMultiSz(devs,&devInfo,SPDRP_COMPATIBLEIDS);

					if(WildCompareHwIds(hwIds,idDevice) ||
						WildCompareHwIds(compatIds,idDevice)) {
							match = TRUE;
					}
				}
				DelMultiSz(hwIds);
				DelMultiSz(compatIds);
			}
		} else {
			match = TRUE;
		}
		if(match) {
			retcode = Callback(devs,&devInfo,devIndex,Context);
			if(retcode) {
				failcode = retcode;
				goto final;
			}
		}
	}

	failcode = EXIT_OK;

final:
	if(devs != INVALID_HANDLE_VALUE) {
		SetupDiDestroyDeviceInfoList(devs);
	}
	return failcode;

}

int ControlCallback(HDEVINFO Devs,PSP_DEVINFO_DATA DevInfo,DWORD Index,LPVOID Context)
/*++

Routine Description:

Callback for use by Enable/Disable/Restart
Invokes DIF_PROPERTYCHANGE with correct parameters
uses SetupDiCallClassInstaller so cannot be done for remote devices
Don't use CM_xxx API's, they bypass class/co-installers and this is bad.

In Enable case, we try global first, and if still disabled, enable local

Arguments:

Devs    )_ uniquely identify the device
DevInfo )
Index    - index of device
Context  - GenericContext

Return Value:

EXIT_xxxx

--*/
{
	SP_DEVINFO_LIST_DETAIL_DATA devInfoListDetail;
    ULONG status = 0;
    ULONG problem = 0;
    BOOL bEnabled = FALSE;
	SP_PROPCHANGE_PARAMS pcp;
	GenericContext *pControlContext = (GenericContext*)Context;
	SP_DEVINSTALL_PARAMS devParams;


	// Check the status of the device to determine what the desired action is:
	// Enable / Disable Toggle
    devInfoListDetail.cbSize = sizeof(devInfoListDetail);
    if((!SetupDiGetDeviceInfoListDetail(Devs,&devInfoListDetail)) ||
            (CM_Get_DevNode_Status_Ex(&status,&problem,DevInfo->DevInst,0,devInfoListDetail.RemoteMachineHandle)!=CR_SUCCESS)) {
        return EXIT_FAIL;
    }
    //
    // check the status/problem codes
    //
    if(status & DN_STARTED) {
        bEnabled = TRUE;
		pControlContext->control = DICS_DISABLE;
    } 
    else if((status & DN_HAS_PROBLEM) && problem == CM_PROB_DISABLED) {
        bEnabled = FALSE;
		pControlContext->control = DICS_ENABLE;
    }
	else
	{
		// The device is in some indeterminite state.  Quit and leave it alone.
		return EXIT_FAIL;
	}

	switch(pControlContext->control) {
case DICS_ENABLE:
	//
	// enable both on global and config-specific profile
	// do global first and see if that succeeded in enabling the device
	// (global enable doesn't mark reboot required if device is still
	// disabled on current config whereas vice-versa isn't true)
	//
	pcp.ClassInstallHeader.cbSize = sizeof(SP_CLASSINSTALL_HEADER);
	pcp.ClassInstallHeader.InstallFunction = DIF_PROPERTYCHANGE;
	pcp.StateChange = pControlContext->control;
	pcp.Scope = DICS_FLAG_GLOBAL;
	pcp.HwProfile = 0;
	//
	// don't worry if this fails, we'll get an error when we try config-
	// specific.
	if(SetupDiSetClassInstallParams(Devs,DevInfo,&pcp.ClassInstallHeader,sizeof(pcp))) {
		SetupDiCallClassInstaller(DIF_PROPERTYCHANGE,Devs,DevInfo);
	}
	//
	// now enable on config-specific
	//
	pcp.ClassInstallHeader.cbSize = sizeof(SP_CLASSINSTALL_HEADER);
	pcp.ClassInstallHeader.InstallFunction = DIF_PROPERTYCHANGE;
	pcp.StateChange = pControlContext->control;
	pcp.Scope = DICS_FLAG_CONFIGSPECIFIC;
	pcp.HwProfile = 0;
	break;

default:
	//
	// operate on config-specific profile
	//
	pcp.ClassInstallHeader.cbSize = sizeof(SP_CLASSINSTALL_HEADER);
	pcp.ClassInstallHeader.InstallFunction = DIF_PROPERTYCHANGE;
	pcp.StateChange = pControlContext->control;
	pcp.Scope = DICS_FLAG_CONFIGSPECIFIC;
	pcp.HwProfile = 0;
	break;

	}

	if(!SetupDiSetClassInstallParams(Devs,DevInfo,&pcp.ClassInstallHeader,sizeof(pcp)) ||
		!SetupDiCallClassInstaller(DIF_PROPERTYCHANGE,Devs,DevInfo)) {
			//
			// failed to invoke DIF_PROPERTYCHANGE
			//
	} else {
		//
		// see if device needs reboot
		//
		devParams.cbSize = sizeof(devParams);
		if(SetupDiGetDeviceInstallParams(Devs,DevInfo,&devParams) && (devParams.Flags & (DI_NEEDRESTART|DI_NEEDREBOOT))) {
			pControlContext->reboot = TRUE;
		} else {
			//
			// appears to have succeeded
			//
		}
		pControlContext->count++;
	}
	return EXIT_OK;
}

int cmdToggle(LPTSTR strClass,LPTSTR strDevice)
/*++

Routine Description:

Get the status of a device to determine the proper toggle behavior.

Arguments:

strClass  - class name
strDevice - wildcard string of device id

Return Value:

EXIT_xxxx (EXIT_REBOOT if reboot is required)

--*/
{
	GenericContext context;
	int failcode = EXIT_FAIL;

	context.control = NULL; // This will be set to DICS_ENABLE DICS_DISABLE based on current state.
	context.reboot = FALSE;
	context.count = 0;
	context.strReboot = _T("");
	context.strSuccess = _T("");
	context.strFail = _T("");
	failcode = EnumerateDevices(DIGCF_PRESENT,strClass,strDevice,ControlCallback,&context);

	if(failcode == EXIT_OK) {

		if(!context.count) {
		} else if(!context.reboot) {
		} else {
			failcode = EXIT_REBOOT;
		}
	}
	return failcode;
}